@@ -6,10 +6,136 @@
 ##
 
 import os, time, torch
-from torch.utils.data import DataLoader
-from .utils import get_tensor_device
+from torch.utils.data import TensorDataset, DataLoader, Dataset
+from .utils import get_tensor_device, list_files, indexOf
 
 
+class TrainHistory:
+	
+	def __init__(self):
+		
+		self.epoch_number = 0
+		self.epoch = {}
+	
+	
+	def add(self, record):
+		
+		"""
+		Add train history from record
+		"""
+		
+		epoch_number = record["epoch_number"]
+		self.epoch_number = epoch_number
+		self.epoch[epoch_number] = {
+			
+			"loss_train": record["loss_train"],
+			"loss_test": record["loss_test"],
+			"acc_train": record["acc_train"],
+			"acc_test": record["acc_test"],
+			"acc_rel": record["acc_rel"],
+			"time": record["time"],
+			"batch_train_iter": record["batch_train_iter"],
+			"batch_test_iter": record["batch_test_iter"],
+			"train_count_iter": record["train_count_iter"],
+			"test_count_iter": record["test_count_iter"],
+			"loss_train_iter": record["loss_train_iter"],
+			"loss_test_iter": record["loss_test_iter"],
+			"acc_train_iter": record["acc_train_iter"],
+			"acc_test_iter": record["acc_test_iter"],
+			
+		}
+	
+	
+	def add_train_status(self, train_status):
+		
+		"""
+		Add train history from train status
+		"""
+		
+		loss_train = train_status.get_loss_train()
+		loss_test = train_status.get_loss_test()
+		acc_train = train_status.get_acc_train()
+		acc_test = train_status.get_acc_test()
+		acc_rel = train_status.get_acc_rel()
+		time = train_status.get_time()
+		
+		self.add( {
+			
+			"epoch_number": train_status.epoch_number,
+			"loss_train": loss_train,
+			"loss_test": loss_test,
+			"acc_train": acc_train,
+			"acc_test": acc_test,
+			"acc_rel": acc_rel,
+			"time": time,
+			"batch_train_iter": train_status.batch_train_iter,
+			"batch_test_iter": train_status.batch_test_iter,
+			"train_count_iter": train_status.train_count_iter,
+			"test_count_iter": train_status.test_count_iter,
+			"loss_train_iter": train_status.loss_train_iter,
+			"loss_test_iter": train_status.loss_test_iter,
+			"acc_train_iter": train_status.acc_train_iter,
+			"acc_test_iter": train_status.acc_test_iter,
+			
+		} )
+	
+	
+	def get_epoch(self, epoch_number):
+		
+		"""
+		Returns epoch by number
+		"""
+		
+		if not(epoch_number in self.epoch):
+			return None
+			
+		return self.epoch[epoch_number]
+	
+	
+	def get_metrics(self, metric_name, with_index=False):
+		
+		"""
+		Returns metrics by name
+		"""
+		
+		res = []
+		
+		for index in self.epoch:
+			
+			if with_index:
+				res.append( (index, self.epoch[index][metric_name]) )
+			else:
+				res.append( self.epoch[index][metric_name] )
+		
+		return res
+		
+	
+	def get_plot(self):
+		
+		"""
+		Returns train history
+		"""
+		
+		import matplotlib.pyplot as plt
+		import numpy as np
+		
+		loss_train = self.get_metrics("loss_train")
+		loss_test = self.get_metrics("loss_test")
+		acc_train = self.get_metrics("acc_train")
+		acc_test = self.get_metrics("acc_test")
+		
+		fig, axs = plt.subplots(2)
+		axs[0].plot( np.multiply(loss_train, 100), label='train loss')
+		axs[0].plot( np.multiply(loss_test, 100), label='test loss')
+		axs[0].legend()
+		axs[1].plot( np.multiply(acc_train, 100), label='train acc')
+		axs[1].plot( np.multiply(acc_test, 100), label='test acc')
+		axs[1].legend()
+		plt.xlabel('Epoch')
+		
+		return plt
+	
+
 class TrainStatus:
 	
 	def __init__(self):
@@ -23,28 +149,30 @@ class TrainStatus:
 		self.acc_train_iter = 0
 		self.acc_test_iter = 0
 		self.epoch_number = 0
+		self.trainer = None
 		self.do_training = True
 		self.train_data_count = 0
 		self.time_start = 0
 		self.time_end = 0
-		self.callbacks = []
 	
 	def set_model(self, model):
 		
 		self.clear()
 		self.model = model
-		self.epoch_number = model.history["epoch_number"]
+		self.epoch_number = model.history.epoch_number
 		
 		if self.epoch_number > 0:
 			
-			self.batch_train_iter = model.history["batch_train_iter"][-1]
-			self.batch_test_iter = model.history["batch_test_iter"][-1]
-			self.train_count_iter = model.history["train_count_iter"][-1]
-			self.test_count_iter = model.history["test_count_iter"][-1]
-			self.loss_train_iter = model.history["loss_train_iter"][-1]
-			self.loss_test_iter = model.history["loss_test_iter"][-1]
-			self.acc_train_iter = model.history["acc_train_iter"][-1]
-			self.acc_test_iter = model.history["acc_test_iter"][-1]
+			record = model.history.get_epoch(self.epoch_number)
+			
+			self.batch_train_iter = record["batch_train_iter"]
+			self.batch_test_iter = record["batch_test_iter"]
+			self.train_count_iter = record["train_count_iter"]
+			self.test_count_iter = record["test_count_iter"]
+			self.loss_train_iter = record["loss_train_iter"]
+			self.loss_test_iter = record["loss_test_iter"]
+			self.acc_train_iter = record["acc_train_iter"]
+			self.acc_test_iter = record["acc_test_iter"]
 	
 	def clear(self):
 		self.clear_iter()
@@ -102,98 +230,21 @@ class TrainStatus:
 	def get_time(self):
 		return self.time_end - self.time_start
 	
-	"""	====================== Events ====================== """
-	
-	def on_start_train(self):
-		"""
-		Start train event
-		"""
-		for callback in self.callbacks:
-			if hasattr(callback, "on_start_train"):
-				callback.on_start_train(self)
-		
-	
-	def on_end_train(self):
-		"""
-		End train event
-		"""
-		for callback in self.callbacks:
-			if hasattr(callback, "on_end_train"):
-				callback.on_end_train(self)
-	
-	
-	def on_start_epoch(self):
-		"""
-		Start epoch event
-		"""
-		for callback in self.callbacks:
-			if hasattr(callback, "on_start_epoch"):
-				callback.on_start_epoch(self)
-	
-	
-	def on_end_epoch(self):
-		
-		"""
-		End epoch event
-		"""
-		
-		self.model.add_train_status_iter(self)
-		
-		for callback in self.callbacks:
-			if hasattr(callback, "on_end_epoch"):
-				callback.on_end_epoch(self)
-	
-	
-	def on_start_batch_train(self, batch_x, batch_y):
-		"""
-		Start train batch event
-		"""
-		for callback in self.callbacks:
-			if hasattr(callback, "on_start_batch_train"):
-				callback.on_start_batch_train(self)
-	
-	
-	def on_end_batch_train(self, batch_x, batch_y):
-		"""
-		End train batch event
-		"""
-		for callback in self.callbacks:
-			if hasattr(callback, "on_end_batch_train"):
-				callback.on_end_batch_train(self)
-	
-	
-	def on_start_batch_test(self, batch_x, batch_y):
-		"""
-		Start test batch event
-		"""
-		for callback in self.callbacks:
-			if hasattr(callback, "on_start_batch_test"):
-				callback.on_start_batch_test(self)
-	
-	
-	def on_end_batch_test(self, batch_x, batch_y):
-		"""
-		End test batch event
-		"""
-		for callback in self.callbacks:
-			if hasattr(callback, "on_end_batch_test"):
-				callback.on_end_batch_test(self)
-	
 
 class TrainVerboseCallback:
 	
 	
-	def on_end_batch_train(self, train_status:TrainStatus):
+	def on_end_batch_train(self, trainer):
 		
-		acc_train = train_status.get_acc_train()
-		loss_train = train_status.get_loss_train()
-		time = train_status.get_time()
+		acc_train = trainer.train_status.get_acc_train()
+		loss_train = trainer.train_status.get_loss_train()
+		time = trainer.train_status.get_time()
 		
 		msg = ("\rStep {epoch_number}, {iter_value}%" +
 			", acc: .{acc}, loss: .{loss}, time: {time}s"
 		).format(
-			epoch_number = train_status.epoch_number,
-			iter_value = round(train_status.get_iter_value() * 100),
+			epoch_number = trainer.train_status.epoch_number,
+			iter_value = round(trainer.train_status.get_iter_value() * 100),
 			loss = str(round(loss_train * 10000)).zfill(4),
 			acc = str(round(acc_train * 100)).zfill(2),
 			time = str(round(time)),
@@ -202,18 +253,18 @@ class TrainVerboseCallback:
 		print (msg, end='')
 	
 	
-	def on_end_epoch(self, train_status:TrainStatus):
+	def on_end_epoch(self, trainer):
 		
 		"""
 		Epoch
 		"""
 		
-		loss_train = train_status.get_loss_train()
-		loss_test = train_status.get_loss_test()
-		acc_train = train_status.get_acc_train()
-		acc_test = train_status.get_acc_test()
-		acc_rel = train_status.get_acc_rel()
-		time = train_status.get_time()
+		loss_train = trainer.train_status.get_loss_train()
+		loss_test = trainer.train_status.get_loss_test()
+		acc_train = trainer.train_status.get_acc_train()
+		acc_test = trainer.train_status.get_acc_test()
+		acc_rel = trainer.train_status.get_acc_rel()
+		time = trainer.train_status.get_time()
 		
 		print ("\r", end='')
 		
@@ -225,17 +276,135 @@ class TrainVerboseCallback:
 			"loss_test: .{loss_test}, " +
 			"time: {time}s, "
 		).format(
-			epoch_number = train_status.epoch_number,
+			epoch_number = trainer.train_status.epoch_number,
 			loss_train = str(round(loss_train * 10000)).zfill(4),
 			loss_test = str(round(loss_test * 10000)).zfill(4),
 			acc_train = str(round(acc_train * 100)).zfill(2),
 			acc_test = str(round(acc_test * 100)).zfill(2),
-			acc_rel = str(round(acc_rel * 100)).zfill(2),
+			acc_rel = str(round(acc_rel * 100) / 100),
 			time = str(round(time)),
 		)
 		
 		print (msg)
 	
+	
+class TrainSaveCallback:
+	
+	
+	def detect_type(self, file_name):
+	
+		import re
+		
+		file_type = ""
+		epoch_index = 0
+		
+		result = re.match(r'^model-(?P<id>[0-9]+)-optimizer\.data$', file_name)
+		if result:
+			return "optimizer", int(result.group("id"))
+		
+		result = re.match(r'^model-(?P<id>[0-9]+)\.data$', file_name)
+		if result:
+			return "model", int(result.group("id"))
+		
+		return file_type, epoch_index
+	
+	
+	def save_epoch_indexes(self, model, epoch_indexes):
+		
+		"""
+		Save epoch by indexes
+		"""
+		
+		model_path = model.get_model_path()
+		files = list_files( model_path )
+		
+		for file_name in files:
+			
+			file_type, epoch_index = self.detect_type(file_name)
+			if file_type in ["model", "optimizer"] and not (epoch_index in epoch_indexes):
+				
+				file_path = os.path.join( model_path, file_name )
+				os.unlink(file_path)
+				
+				print ("File " + file_name + " is deleted")
+				
+	
+	
+	def get_the_best_epoch(self, model, epoch_count=5, indexes=False):
+		
+		"""
+		Returns teh best epoch
+		"""
+		
+		acc_test = model.history.get_metrics("acc_test", with_index=True)
+		
+		def get_key(item):
+			return item[1]
+
+		acc_test.sort(key=get_key, reverse=True)
+		
+		res = []
+		res_count = 0
+		acc_test_len = len(acc_test)
+		acc_test_last = 1000
+		for index in range(acc_test_len):
+			
+			res.append( acc_test[index] )
+			
+			if acc_test_last != acc_test[index][1]:
+				res_count = res_count + 1
+			
+			acc_test_last = acc_test[index][1]
+			
+			if res_count > epoch_count:
+				break
+		
+		if not indexes:
+			return res
+			
+		res_indexes = []
+		for index in range(len(res)):
+			res_indexes.append( res[index][0] )
+			
+		return res_indexes
+	
+	
+	def save_the_best_epoch(self, model, epoch_count=5):
+		
+		"""
+		Save the best models
+		"""
+		
+		if model.history.epoch_number > 0 and epoch_count > 0:
+			
+			epoch_indexes = self.get_the_best_epoch(model, epoch_count, indexes=True)
+			epoch_indexes.append( model.history.epoch_number )
+			
+			self.save_epoch_indexes(model, epoch_indexes)
+	
+	
+	def on_end_epoch(self, trainer):
+		
+		"""
+		On epoch end
+		"""
+		
+		trainer.model.save(
+			save_epoch=trainer.save_epoch
+		)
+		
+		if trainer.save_epoch:
+			trainer.model.save_optimizer(trainer.optimizer)
+			
+			"""
+			Save best epoch
+			"""
+			self.save_the_best_epoch(
+				trainer.model,
+				epoch_count=trainer.save_epoch_count
+			)
+		
+		
 
 class Trainer:
 	
@@ -244,6 +413,7 @@ class Trainer:
 		self.model = model
 		
 		self.train_status = TrainStatus()
+		self.train_status.trainer = self
 		self.train_status.set_model(model)
 		
 		self.train_loader = None
@@ -251,6 +421,7 @@ class Trainer:
 		self.train_dataset = None
 		self.test_dataset = None
 		self.batch_size = 64
+		self.lr = 1e-3
 		
 		self.optimizer = None
 		self.loss = None
@@ -267,14 +438,19 @@ class Trainer:
 		self.train_dataset = kwargs["train_dataset"] if "train_dataset" in kwargs else False
 		self.test_dataset = kwargs["test_dataset"] if "test_dataset" in kwargs else False
 		self.tensor_device = kwargs["tensor_device"] if "tensor_device" in kwargs else None
+		self.save_epoch = kwargs["save_epoch"] if "save_epoch" in kwargs else True
+		self.save_epoch_count = kwargs["save_epoch_count"] if "save_epoch_count" in kwargs else 5
 		
 		self._check_is_trained = kwargs["check_is_trained"] \
 			if "check_is_trained" in kwargs else None
 		
 		if "callbacks" in kwargs:
-			self.train_status.callbacks = kwargs["callbacks"]
+			self.callbacks = kwargs["callbacks"]
 		else:
-			self.train_status.callbacks = [ TrainVerboseCallback() ]
+			self.callbacks = [
+				TrainVerboseCallback(),
+				TrainSaveCallback(),
+			]
 		
 	
 	def check_is_trained(self):
@@ -306,7 +482,75 @@ class Trainer:
 		
 		return False
 	
+	
+	"""	====================== Events ====================== """
+	
+	def on_start_train(self):
 		
+		"""
+		Start train event
+		"""
+		
+		for callback in self.callbacks:
+			if hasattr(callback, "on_start_train"):
+				callback.on_start_train(self)
+		
+	
+	def on_start_epoch(self):
+		
+		"""
+		Start epoch event
+		"""
+		
+		for callback in self.callbacks:
+			if hasattr(callback, "on_start_epoch"):
+				callback.on_start_epoch(self)
+	
+	
+	def on_start_batch_train(self, batch_x, batch_y):
+		
+		"""
+		Start train batch event
+		"""
+		
+		for callback in self.callbacks:
+			if hasattr(callback, "on_start_batch_train"):
+				callback.on_start_batch_train(self)
+	
+	
+	def on_end_batch_train(self, batch_x, batch_y):
+		
+		"""
+		End train batch event
+		"""
+		
+		for callback in self.callbacks:
+			if hasattr(callback, "on_end_batch_train"):
+				callback.on_end_batch_train(self)
+	
+	
+	def on_start_batch_test(self, batch_x, batch_y):
+		
+		"""
+		Start test batch event
+		"""
+		
+		for callback in self.callbacks:
+			if hasattr(callback, "on_start_batch_test"):
+				callback.on_start_batch_test(self)
+	
+	
+	def on_end_batch_test(self, batch_x, batch_y):
+		
+		"""
+		End test batch event
+		"""
+		
+		for callback in self.callbacks:
+			if hasattr(callback, "on_end_batch_test"):
+				callback.on_end_batch_test(self)
+	
+	
 	def on_end_epoch(self, **kwargs):
 		
 		"""
@@ -318,9 +562,24 @@ class Trainer:
 		if self._is_trained:
 			self.stop_training()
 		
-		self.model.save()
+		self.model.history.add_train_status(self.train_status)
 		
+		for callback in self.callbacks:
+			if hasattr(callback, "on_end_epoch"):
+				callback.on_end_epoch(self)
 		
+	
+	def on_end_train(self):
+		
+		"""
+		End train event
+		"""
+		
+		for callback in self.callbacks:
+			if hasattr(callback, "on_end_train"):
+				callback.on_end_train(self)
+	
+	
 	def stop_training(self):
 		
 		"""
@@ -340,8 +599,11 @@ class Trainer:
 		
 		# Adam optimizer
 		if self.optimizer is None:
-			self.optimizer = torch.optim.Adam(model.module.parameters(), lr=3e-4)
+			self.optimizer = torch.optim.Adam(model.module.parameters(), lr=self.lr)
 		
+		if self.save_epoch:
+			model.load_optimizer(self.optimizer)
+			
 		# Mean squared error
 		if self.loss is None:
 			self.loss = torch.nn.MSELoss()
@@ -373,7 +635,7 @@ class Trainer:
 		train_status = self.train_status
 		train_status.do_training = True
 		train_status.train_data_count = len(self.train_dataset)
-		train_status.on_start_train()
+		self.on_start_train()
 		
 		try:
 		
@@ -382,7 +644,7 @@ class Trainer:
 				train_status.clear_iter()
 				train_status.epoch_number = train_status.epoch_number + 1
 				train_status.time_start = time.time()
-				train_status.on_start_epoch()
+				self.on_start_epoch()
 				
 				module.train()
 				
@@ -393,7 +655,7 @@ class Trainer:
 					batch_y = batch_y.to(tensor_device)
 					batch_x, batch_y = model.convert_batch(x=batch_x, y=batch_y)
 					
-					train_status.on_start_batch_train(batch_x, batch_y)
+					self.on_start_batch_train(batch_x, batch_y)
 					
 					# Predict model
 					batch_predict = module(batch_x)
@@ -420,7 +682,7 @@ class Trainer:
 					train_status.train_count_iter = train_status.train_count_iter + batch_x.shape[0]
 					
 					train_status.time_end = time.time()
-					train_status.on_end_batch_train(batch_x, batch_y)
+					self.on_end_batch_train(batch_x, batch_y)
 					
 					del batch_x, batch_y
 					
@@ -437,7 +699,7 @@ class Trainer:
 					batch_y = batch_y.to(tensor_device)
 					batch_x, batch_y = model.convert_batch(x=batch_x, y=batch_y)
 					
-					train_status.on_start_batch_test(batch_x, batch_y)
+					self.on_start_batch_test(batch_x, batch_y)
 					
 					# Predict model
 					batch_predict = module(batch_x)
@@ -459,7 +721,7 @@ class Trainer:
 					train_status.test_count_iter = train_status.test_count_iter + batch_x.shape[0]
 					
 					train_status.time_end = time.time()
-					train_status.on_end_batch_test(batch_x, batch_y)
+					self.on_end_batch_test(batch_x, batch_y)
 					
 					del batch_x, batch_y
 					
@@ -469,7 +731,6 @@ class Trainer:
 				
 				# Epoch callback
 				train_status.time_end = time.time()
-				train_status.on_end_epoch()
 				self.on_end_epoch()
 				
 				if not train_status.do_training:
@@ -483,7 +744,7 @@ class Trainer:
 			
 			pass
 		
-		train_status.on_end_train()
+		self.on_end_train()
 	
 
 def do_train(model, *args, **kwargs):
@@ -499,4 +760,38 @@ def do_train(model, *args, **kwargs):
 		print ("Train model " + str(model.model_name))
 		trainer.train()
 	
-	return trainer
\ No newline at end of file
+	return trainer
+
+
+class FilesListDataset(Dataset):
+	
+	def __init__(self, files, folder_path="", transform=None, get_tensor_from_answer=None):
+		
+		self.files = files
+		self.folder_path = folder_path
+		self.transform = transform
+		self.get_tensor_from_answer = get_tensor_from_answer
+	
+	
+	def __getitem__(self, index):
+		
+		file_path = ""
+		answer = None
+		
+		if self.get_tensor_from_answer is None:
+			file_path = self.files[index]
+		else:
+			file_path, answer = self.files[index]
+		
+		if self.folder_path != "":
+			file_path = os.path.join(self.folder_path, file_path)
+		
+		tensor = self.transform(file_path)
+		
+		if self.get_tensor_from_answer is None:
+			answer = self.get_tensor_from_answer(answer)
+		
+		return ( tensor, answer )
+	
+	def __len__(self):
+		return len(self.files)
